@@ -1,11 +1,10 @@
-// config.js (c) 2010-2020 Loren West and other contributors
+// config.js (c) 2010-2022 Loren West and other contributors
 // May be freely distributed under the MIT license.
 // For further details and documentation:
 // http://lorenwest.github.com/node-config
 
 // Dependencies
-var deferConfig = require('../defer').deferConfig,
-    DeferredConfig = require('../defer').DeferredConfig,
+var DeferredConfig = require('../defer').DeferredConfig,
     RawConfig = require('../raw').RawConfig,
     Parser = require('../parser'),
     Utils = require('util'),
@@ -14,12 +13,10 @@ var deferConfig = require('../defer').deferConfig,
 
 // Static members
 var DEFAULT_CLONE_DEPTH = 20,
-    NODE_CONFIG, CONFIG_DIR, RUNTIME_JSON_FILENAME, NODE_ENV, APP_INSTANCE,
-    HOST, HOSTNAME, ALLOW_CONFIG_MUTATIONS, CONFIG_SKIP_GITCRYPT, NODE_ENV_VAR_NAME,
+    CONFIG_DIR, NODE_ENV, APP_INSTANCE,
+    CONFIG_SKIP_GITCRYPT, NODE_ENV_VAR_NAME,
     NODE_CONFIG_PARSER,
     env = {},
-    privateUtil = {},
-    deprecationWarnings = {},
     configSources = [],          // Configuration sources - array of {name, original, parsed}
     checkMutability = true,      // Check for mutability/immutability on first get
     gitCryptTestRegex = /^.GITCRYPT/; // regular expression to test for gitcrypt files.
@@ -410,19 +407,28 @@ util.makeImmutable = function(object, property, value) {
         value: Object.freeze(value)
       });
     } else {
+      // Call recursively if an object.
+      if (util.isObject(value)) {
+        // Create a proxy, to capture user updates of configuration options, and throw an exception for awareness, as per:
+        // https://github.com/lorenwest/node-config/issues/514
+        value = new Proxy(util.makeImmutable(value), {
+          set (target, name) {
+            const message = (Reflect.has(target, name) ? 'update' : 'add');
+            // Notify the user.
+            throw Error(`Can not ${message} runtime configuration property: "${name}". Configuration objects are immutable unless ALLOW_CONFIG_MUTATIONS is set.`)
+          }
+        })
+      }
+
       Object.defineProperty(object, propertyName, {
         value: value,
         writable : false,
         configurable: false
       });
 
-      // Ensure new properties can not be added.
-      Object.preventExtensions(object)
-
-      // Call recursively if an object.
-      if (util.isObject(value)) {
-        util.makeImmutable(value);
-      }
+      // Ensure new properties can not be added, as per:
+      // https://github.com/lorenwest/node-config/issues/505
+      Object.preventExtensions(object[propertyName])
     }
   }
 
@@ -562,25 +568,21 @@ util.loadFileConfigs = function(configDir, options) {
   // Split files name, for loading multiple files.
   NODE_ENV = NODE_ENV.split(',');
 
-  CONFIG_DIR = configDir || util.initParam('NODE_CONFIG_DIR', Path.join( process.cwd(), 'config') );
-  if (CONFIG_DIR.indexOf('.') === 0) {
-    CONFIG_DIR = Path.join(process.cwd() , CONFIG_DIR);
-  }
+  var dir = configDir || util.initParam('NODE_CONFIG_DIR', Path.join( process.cwd(), 'config') );
+  dir = _toAbsolutePath(dir);
 
   APP_INSTANCE = util.initParam('NODE_APP_INSTANCE');
-  HOST = util.initParam('HOST');
-  HOSTNAME = util.initParam('HOSTNAME');
   CONFIG_SKIP_GITCRYPT = util.initParam('CONFIG_SKIP_GITCRYPT');
 
   // This is for backward compatibility
-  RUNTIME_JSON_FILENAME = util.initParam('NODE_CONFIG_RUNTIME_JSON', Path.join(CONFIG_DIR , 'runtime.json') );
+  var runtimeFilename = util.initParam('NODE_CONFIG_RUNTIME_JSON', Path.join(dir , 'runtime.json') );
 
   NODE_CONFIG_PARSER = util.initParam('NODE_CONFIG_PARSER');
   if (NODE_CONFIG_PARSER) {
     try {
       var parserModule = Path.isAbsolute(NODE_CONFIG_PARSER)
         ? NODE_CONFIG_PARSER
-        : Path.join(CONFIG_DIR, NODE_CONFIG_PARSER);
+        : Path.join(dir, NODE_CONFIG_PARSER);
       Parser = require(parserModule);
     }
     catch (e) {
@@ -589,6 +591,9 @@ util.loadFileConfigs = function(configDir, options) {
     }
   }
 
+  var HOST = util.initParam('HOST');
+  var HOSTNAME = util.initParam('HOSTNAME');
+
   // Determine the host name from the OS module, $HOST, or $HOSTNAME
   // Remove any . appendages, and default to null if not set
   try {
@@ -639,7 +644,7 @@ util.loadFileConfigs = function(configDir, options) {
     });
   });
 
-  var locatedFiles = util.locateMatchingFiles(CONFIG_DIR, allowedFiles);
+  var locatedFiles = util.locateMatchingFiles(dir, allowedFiles);
   locatedFiles.forEach(function(fullFilename) {
     var configObj = util.parseFile(fullFilename, options);
     if (configObj) {
@@ -651,6 +656,9 @@ util.loadFileConfigs = function(configDir, options) {
   // NODE_CONFIG only applies to the base config
   if (!configDir) {
     var envConfig = {};
+
+    CONFIG_DIR = dir;
+
     if (process.env.NODE_CONFIG) {
       try {
         envConfig = JSON.parse(process.env.NODE_CONFIG);
@@ -690,11 +698,11 @@ util.loadFileConfigs = function(configDir, options) {
   }
 
   // Override with environment variables if there is a custom-environment-variables.EXT mapping file
-  var customEnvVars = util.getCustomEnvVars(CONFIG_DIR, extNames);
+  var customEnvVars = util.getCustomEnvVars(dir, extNames);
   util.extendDeep(config, customEnvVars);
 
   // Extend the original config with the contents of runtime.json (backwards compatibility)
-  var runtimeJson = util.parseFile(RUNTIME_JSON_FILENAME) || {};
+  var runtimeJson = util.parseFile(runtimeFilename) || {};
   util.extendDeep(config, runtimeJson);
 
   util.resolveDeferredConfigs(config);
@@ -718,6 +726,7 @@ util.locateMatchingFiles = function(configDirs, allowedFiles) {
   return configDirs.split(Path.delimiter)
     .reduce(function(files, configDir) {
       if (configDir) {
+        configDir = _toAbsolutePath(configDir);
         try {
           FileSystem.readdirSync(configDir).forEach(function(file) {
             if (allowedFiles[file]) {
@@ -1118,18 +1127,18 @@ util.substituteDeep = function (substitutionMap, variables) {
  *
  * @protected
  * @method getCustomEnvVars
- * @param CONFIG_DIR {string} - the passed configuration directory
+ * @param configDir {string} - the passed configuration directory
  * @param extNames {Array[string]} - acceptable configuration file extension names.
  * @returns {object} - mapped environment variables or {} if there are none
  */
-util.getCustomEnvVars = function (CONFIG_DIR, extNames) {
+util.getCustomEnvVars = function (configDir, extNames) {
   var result = {};
   var resolutionIndex = 1;
   var allowedFiles = {};
   extNames.forEach(function (extName) {
     allowedFiles['custom-environment-variables' + '.' + extName] = resolutionIndex++;
   });
-  var locatedFiles = util.locateMatchingFiles(CONFIG_DIR, allowedFiles);
+  var locatedFiles = util.locateMatchingFiles(configDir, allowedFiles);
   locatedFiles.forEach(function (fullFilename) {
     var configObj = util.parseFile(fullFilename);
     if (configObj) {
@@ -1476,7 +1485,7 @@ util.runStrictnessChecks = function (config) {
   function _warnOrThrow (msg) {
     var beStrict = process.env.NODE_CONFIG_STRICT_MODE;
     var prefix = beStrict ? 'FATAL: ' : 'WARNING: ';
-    var seeURL = 'See https://github.com/lorenwest/node-config/wiki/Strict-Mode';
+    var seeURL = 'See https://github.com/node-config/node-config/wiki/Strict-Mode';
 
     console.error(prefix+msg);
     console.error(prefix+seeURL);
@@ -1488,6 +1497,15 @@ util.runStrictnessChecks = function (config) {
   }
 };
 
+// Helper functions shared accross object members
+function _toAbsolutePath (configDir) {
+  if (configDir.indexOf('.') === 0) {
+    return Path.join(process.cwd(), configDir);
+  }
+
+  return configDir;
+}
+
 // Instantiate and export the configuration
 var config = module.exports = new Config();
 
